@@ -20,6 +20,7 @@ import jax
 from jax import abstract_arrays
 from jax import ad_util
 from jax import api
+from jax import config
 from jax import core
 from jax import custom_derivatives
 from jax import dtypes
@@ -153,6 +154,11 @@ def convert(fun, with_gradient=True):
   api._check_callable(fun)
 
   def converted_fun(*args: TfVal) -> TfVal:
+    # TODO: is there a better way to check if we are inside a transformation?
+    if config.omnistaging_enabled and not core.trace_state_clean():
+      raise ValueError("convert must be used outside all JAX transformations."
+                       + f"Trace state: {core.thread_local_state.trace_state}")
+
     # This function may take pytrees of TfVals. We can only set
     # tf.custom_gradient on functions that take a flat argument list.
     args_flat, in_tree = tree_util.tree_flatten((args, {}))
@@ -211,7 +217,8 @@ def convert(fun, with_gradient=True):
 
 def _interpret_fun(fun: lu.WrappedFun,
                    in_vals: Sequence[TfValOrUnit]) -> Sequence[TfValOrUnit]:
-  with core.new_main(TensorFlowTrace) as main:
+  new_main = core.new_base_main if config.omnistaging_enabled else core.new_main
+  with new_main(TensorFlowTrace) as main:  # type: ignore
     fun = _interpret_subtrace(fun, main)
     out_vals: Sequence[TfValOrUnit] = fun.call_wrapped(*in_vals)
     del main
@@ -312,18 +319,37 @@ class TensorFlowTracer(core.Tracer):
 
 
 class TensorFlowTrace(core.Trace):
-  """Trace class that underlies the jax2tf transformation."""
+  """Trace class that underlies the jax2tf transformation.
+
+  We are going to ensure that jax2tf.convert is never nested inside other
+  transformations. This is sufficient for intended use cases (converting
+  fully-transformed JAX code). It also simplifies our job because we do not have
+  to handle situations where we apply primitives on a mix of TF values and
+  JAX tracers from an outer transformation. E.g., for addition both the TF values
+  and the JAX tracers have an override and they get confused if they see values
+  from the other world.
+
+  Hence a TFT trace does not interact with non-TFT traces at lower-level. For
+  higher-order control-flow primitives we invoke recursively
+  _interpret_fun on the body of the conditional, which will create a nested TFT.
+
+  We do want to allow transformations nested inside a TensorFlowTrace (TFT), but
+  those will introduce their own MainTrace, and any operations involving those
+  will be done on those traces, i.e., not a concern for TFT.
+  """
   def pure(self, val: TfValOrUnit):
     """Lifts a non-Tracer into the TensorFlowTrace."""
     return TensorFlowTracer(self, val)
 
   def lift(self, val: core.Tracer):
-    """Lifts a core.Tracer from a lower-level main into the TensorFlowTrace."""
-    # TODO(necula): this should never be needed
-    return TensorFlowTracer(self, val)
+    # This would be called when we need to raise a tracer from a lower-level
+    # main into the TensorFlowTrace. Since the TensorFlowTrace is never nested
+    # inside another transform, there are no lower-level main traces.
+    assert False
 
   def sublift(self, val: TensorFlowTracer):
-    # TODO(necula): this should never be needed
+    # This is called when we need to raise a tracer from the same master,
+    # but a lower sublevel. This could come from a nested jit.
     return TensorFlowTracer(self, val.val)
 
   def process_primitive(self, primitive: core.Primitive,
@@ -344,7 +370,7 @@ class TensorFlowTrace(core.Trace):
         *[t.aval for t in tracers], **params)
       if primitive.multiple_results:
         for o, expected_aval in zip(out, expected_out_aval):  # type: ignore
-          assert o.aval == expected_aval, (
+          assert o.aval.strip_weak_type() == expected_aval.strip_weak_type(), (
             f"{primitive}: out.aval = {o.aval}; expected {expected_aval}")
       else:
         assert out.aval == expected_out_aval, (  # type: ignore
@@ -380,13 +406,15 @@ class TensorFlowTrace(core.Trace):
   def get_primitive_impl(self, p):
     try:
       return tf_impl[p]
-    except KeyError:
+    except KeyError as err:
       msg = "TensorFlow interpretation rule for '{}' not implemented"
-      raise NotImplementedError(msg.format(p))
+      raise NotImplementedError(msg.format(p)) from err
 
 def to_tf_dtype(jax_dtype):
   if jax_dtype == jnp.bfloat16:
     return tf.bfloat16
+  elif jax_dtype == dtypes.float0:
+    return tf.float32
   else:
     return tf.dtypes.as_dtype(jax_dtype)
 
@@ -531,7 +559,7 @@ _UNSIGNED_TO_SIGNED_TABLE = {u: s for s, u in _SIGNED_TO_UNSIGNED_TABLE.items()}
 
 # Note: Bitwise operations only yield identical results on unsigned integers!
 # pylint: disable=protected-access
-def _shift_right_arithmetic(x, y):
+def _shift_right_arithmetic_raw(x, y):
   if x.dtype.is_unsigned:
     assert x.dtype == y.dtype
     orig_dtype = x.dtype
@@ -542,9 +570,19 @@ def _shift_right_arithmetic(x, y):
     return tf.cast(res, orig_dtype)
   else:
     return tf.bitwise.right_shift(x, y)
-tf_impl[lax.shift_right_arithmetic_p] = _shift_right_arithmetic
 
-def _shift_right_logical(x, y):
+def _shift_right_arithmetic(x, y):
+  # TF shift is "implementation defined" if the shift amount is negative
+  # or larger or equal to the size of the value. We implement the XLA
+  # semantics to return the shift by the max value (x_bits - 1).
+  # TODO: it is likely better to add XlaOps for shifts
+  x_bits = 8 * x.dtype.size
+  clamp_y = tf.where(_shift_in_bounds(x, y), y, x_bits - 1)
+  return _shift_right_arithmetic_raw(x, clamp_y)
+
+tf_impl[lax.shift_right_arithmetic_p] = wrap_binary_op(_shift_right_arithmetic)
+
+def _shift_right_logical_raw(x, y):
   if x.dtype.is_unsigned:
     return tf.bitwise.right_shift(x, y)
   else:
@@ -555,9 +593,38 @@ def _shift_right_logical(x, y):
     y = tf.cast(y, unsigned_dtype)
     res = tf.bitwise.right_shift(x, y)
     return tf.cast(res, orig_dtype)
-tf_impl[lax.shift_right_logical_p] = _shift_right_logical
 
-tf_impl[lax.shift_left_p] = tf.bitwise.left_shift
+def _shift_right_logical(x, y):
+  # TF shift is "implementation defined" if the shift amount is negative
+  # or larger or equal to the size of the value. We implement the XLA semantics
+  # to return 0.
+  # TODO: it is likely better to add XlaOps for shifts
+  return tf.where(_shift_in_bounds(x, y),
+                  _shift_right_logical_raw(x, y),
+                  tf.zeros_like(x))
+
+tf_impl[lax.shift_right_logical_p] = wrap_binary_op(_shift_right_logical)
+
+def _shift_left(x, y):
+  # TF shift is "implementation defined" if the shift amount is negative
+  # or larger or equal to the size of the value. We implement the XLA semantics
+  # to return 0.
+  # TODO: it is likely better to add XlaOps for shifts
+  return tf.where(_shift_in_bounds(x, y),
+                  tf.bitwise.left_shift(x, y),
+                  tf.zeros_like(x))
+
+tf_impl[lax.shift_left_p] = wrap_binary_op(_shift_left)
+
+def _shift_in_bounds(x: TfVal, y: TfVal) -> TfVal:
+  # Return the TF expression for when y is within bounds (0 <= y < |x|)
+  x_bits = 8 * x.dtype.size
+  # TF does not have comparisons for uint16 and uint32 (despite what the
+  # documentation says)
+  y_comp = tf.cast(y, _UNSIGNED_TO_SIGNED_TABLE[y.dtype]) if y.dtype.is_unsigned else y
+  y_lt_x_bits = tf.math.less(y_comp, x_bits)
+  y_ge_0 = tf.math.greater_equal(y_comp, 0)
+  return tf.logical_and(y_lt_x_bits, y_ge_0)
 
 def _not(x):
   """Computes bitwise not with support for booleans.
@@ -673,6 +740,128 @@ def _conv_general_precision_config_proto(precision):
   proto.operand_precision.append(int(precision))
   return proto
 
+# _try_tf_conv returns a Tensor when it succeeds, or a string describing why
+# it did not succeed otherwise.
+def _try_tf_conv(lhs, rhs, window_strides, padding, lhs_dilation, rhs_dilation,
+                 dimension_numbers, feature_group_count, batch_group_count,
+                 out_shape) -> Union[str, TfVal]:
+  # TODO(bchetioui): this function is not exhaustive wrt which convolution cases
+  # can be translated into TF primitives. Further investigation is needed to
+  # fully flesh it out.
+  if not lhs.dtype in [tf.float16, tf.float32, tf.float64]:
+    return f"tf.nn.convolution is not supported for dtype {lhs.dtype}"
+  if feature_group_count != 1:
+    return "tf.nn.convolution does not support grouped convolutions"
+  # TODO(bchetioui): is there something to do with batch_group_count?
+  if batch_group_count != 1:
+    return "Unimplemented support for batch_group_count != 1"
+  nb_spatial_dimensions = len(lhs.shape) - 2
+  # TF can only deal with 1D, 2D and 3D convolution
+  if nb_spatial_dimensions < 1 or nb_spatial_dimensions > 3:
+    return ("TensorFlow can only handle convolutions with 1, 2, or 3 "
+            "spatial dimensions")
+  # TODO(bchetioui): handle different stride cases
+  if list(window_strides) != [1] * nb_spatial_dimensions:
+    return ("Unimplemented support for window_strides != "
+            f"{tuple([1] * nb_spatial_dimensions)}")
+
+  success = lambda res: (res, None)
+  failure = lambda msg: (None, msg)
+
+  def convert_padding():
+    # TODO(bchetioui): in this instance, we can not use padtype_to_pads as
+    # string padding is not implemented for transposed convolution.
+    if list(lhs_dilation) != [1] * nb_spatial_dimensions:
+      return failure("Padding conversion is not supported for transposed "
+                     "convolution.")
+    lhs_perm, rhs_perm, _ = dimension_numbers
+    effective_rhs_shape = [(k-1) * r + 1 for k, r in
+                           zip(np.take(rhs.shape, rhs_perm)[2:], rhs_dilation)]
+    lhs_shape = np.take(lhs.shape, lhs_perm)[2:]
+    # TF only allows 'VALID' and 'SAME' padding
+    for pad_str in ['VALID', 'SAME']:
+      gen_padding = lax.padtype_to_pads(
+          lhs_shape, effective_rhs_shape, window_strides, pad_str)
+      if list(gen_padding) == list(padding):
+        return success(pad_str)
+    return failure("Input padding not supported in TensorFlow.")
+
+  def convert_dim_nums():
+    lhs_spec, rhs_spec, out_spec = dimension_numbers
+    # TF only allows filters with shape:
+    # spatial_filter_shape + [in_channels, out_channels]. In JAX however,
+    # rhs_spec is represented as a tuple containing the following:
+    # [out_channels, in_channels] + spatial_filter_shape.
+    supported_rhs_shape = ([nb_spatial_dimensions + 1, nb_spatial_dimensions] +
+                           list(range(nb_spatial_dimensions)))
+    if list(rhs_spec) != supported_rhs_shape:
+      return failure("Input filter (RHS) shape format not supported in "
+                     "TensorFlow")
+    # TF only supports same LHS and output data format
+    if lhs_spec != out_spec:
+      return failure("TensorFlow requires the same data format for LHS and "
+                     "output.")
+    # Alphabet extracted from the documentation of tf.conv{1,2,3}d
+    spatial_dim_alphabet = 'DHW'[-nb_spatial_dimensions:]
+    # TF only supports the following data formats:
+    # - [batch_size, in_channels] + input_spatial_shape
+
+    # TODO(bchetioui): TF currently does not support the above on CPU. To avoid
+    # failing on this platform, this path is commented out for now.
+    #if list(lhs_spec) == list(range(len(lhs_spec))):
+    #  return "NC" + spatial_dim_alphabet
+
+    # - [batch_size] + input_spatial_shape + [in_channels]
+    if list(lhs_spec) == ([0, len(lhs_spec) - 1] +
+                          list(range(1, len(lhs_spec) - 1))):
+      return success("N" + spatial_dim_alphabet + "C")
+    return failure("Data format is unsupported by TensorFlow")
+
+  def convert_dilation_and_compute_result(tf_padding, tf_dim_nums):
+    no_dilation = [1] * nb_spatial_dimensions
+      # TODO(bchetioui): is there a generic way to do a transposed atrous
+      # convolution in TensorFlow?
+    if not (list(lhs_dilation) == no_dilation or
+            list(rhs_dilation) == no_dilation):
+      return "Both LHS and RHS dilations are set"
+    # This is a non-dilated or atrous convolution
+    if list(lhs_dilation) == no_dilation:
+      return tf.nn.convolution(
+          lhs, rhs, strides=window_strides, padding=tf_padding,
+          data_format=tf_dim_nums, dilations=rhs_dilation)
+    # TODO(bchetioui): the below path is unreachable for now, as passing a lhs
+    # dilation to this function will result in convert_padding returning None
+    # systematically. This must be investigated further.
+    # Dilation of the LHS is transposed convolution
+    return tf.nn.conv_transpose(
+        lhs, rhs, out_shape, window_strides, padding=tf_padding,
+        data_format=tf_dim_nums, dilations=lhs_dilation)
+
+  tf_padding, error = convert_padding()
+  if tf_padding is None:
+    return error
+  tf_dim_nums, error = convert_dim_nums()
+  if tf_dim_nums is None:
+    return error
+  return convert_dilation_and_compute_result(tf_padding, tf_dim_nums)
+
+# TODO(bchetioui): enabling this flag permits using a conversion path purely
+# based on TF (and not XLA) for _conv_general_dilated in cases when it is
+# possible. It is disabled by default due to a so far unknown bug when running
+# a test in compiled mode. The test that fails is
+#
+# test_conv_general_dilated_tf_conversion_path_3d_lhs=float32[1,4,28,28,1]_rhs=float32[2,3,3,1,16]_windowstrides=(1,1,1)_padding=VALID_lhsdilation=(1,1,1)_rhsdilation=(1,1,2)_dimensionnumbers=('NDHWC','DHWIO','NDHWC')_featuregroupcount=1_batchgroupcount=1_precision=None
+#
+# with the following assertion error in TensorFlowTrace.process_primitive:
+#
+# AssertionError: conv_general_dilated: out.aval = ShapedArray(float32[1,3,24,26,16]); expected ShapedArray(float32[1,3,26,24,16])
+#
+# Deactivating this assertion is enough to pass the test, which suggests that
+# the end shape is indeed the correct one (i.e. (1,3,26,24,16)). Further
+# investigation is required to really understand this behavior, which we have
+# not managed to reproduce as a pure TF test.
+ENABLE_TF_CONVOLUTION = False
+
 def _conv_general_dilated(lhs, rhs, window_strides, padding, lhs_dilation,
                           rhs_dilation, dimension_numbers, feature_group_count,
                           batch_group_count, lhs_shape, rhs_shape, precision):
@@ -681,6 +870,15 @@ def _conv_general_dilated(lhs, rhs, window_strides, padding, lhs_dilation,
       lhs, rhs, window_strides, padding, lhs_dilation, rhs_dilation,
       dimension_numbers, feature_group_count, batch_group_count, lhs_shape,
       rhs_shape, precision)
+
+  if ENABLE_TF_CONVOLUTION:
+    info_or_result = _try_tf_conv(
+        lhs, rhs, window_strides, padding, lhs_dilation, rhs_dilation,
+        dimension_numbers, feature_group_count, batch_group_count, out_shape
+    )
+    if not isinstance(info_or_result, str):
+      return info_or_result
+
   dnums_proto = _conv_general_dimension_numbers_proto(dimension_numbers)
   precision_config_proto = _conv_general_precision_config_proto(precision)
   assert batch_group_count == 1  # TODO(phawkins): implement batch_group_count
@@ -701,6 +899,54 @@ def _dot_general(lhs, rhs, dimension_numbers, precision):
   """Implementation of lax.dot_general_p in terms of tf.linalg.einsum."""
   del precision
   (lhs_contracting, rhs_contracting), (lhs_batch, rhs_batch) = dimension_numbers
+  lhs_dim, rhs_dim = len(lhs.shape), len(rhs.shape)
+  # This condition ensures that:
+  # 1) the considered dtype is not tf.bfloat16/tf.int32, which are supported by
+  #    tf.linalg.einsum but not by tf.linalg.matmul;
+  # 2) the batch dimensions are ordered in the same way in lhs and rhs (this is
+  #    not strictly necessary, but we would have to reshape the array if that
+  #    were not the case;
+  # 3) lhs and rhs have the same number of dimensions +/- 1
+  # 4) the number of non-batch dimensions in both tensors is either 1 or 2
+  # 5) the contracting dimensions are consistent with those of a classic
+  #    matrix/matrix, vector/matrix or matrix/vector multiplication.
+  if (not lhs.dtype in [tf.bfloat16, tf.int32]
+      and lhs_batch == rhs_batch == tuple(range(len(lhs_batch)))
+      and lhs_dim - rhs_dim in [-1, 0, 1]
+      and 1 <= lhs_dim - len(lhs_batch) <= 2
+      and 1 <= rhs_dim - len(rhs_batch) <= 2
+      and lhs_contracting == (len(lhs.shape) - 1,)
+      and rhs_contracting == (len(lhs_batch),)):
+        # All the inputs to tf.linalg.matmul must have 2 inner dimensions,
+        # after their batch dimensions, so we need to expand the dimensions
+        # appropriately. We can get to this branch with three combinations of
+        # inner shapes:
+        # - lhs.inner_shape == [a, b], rhs.inner_shape == [b, c]
+        #   - in this case, the resulting inner shape is [a, c];
+        # - lhs.inner_shape == [b]   , rhs.inner_shape == [b, c]
+        #   - in this case, we need to expand lhs to [1, b], and the resulting
+        #     shape is [c]. We need to squeeze the result of tf.linalg.matmul
+        #     as it will have shape [1, c];
+        # - lhs.shape == [batch] + [a, b], rhs.shape == [batch] + [b]
+        #   - in this case, we need to expand rhs to [b, 1], and the resulting
+        #     shape is [a]. We need to squeeze the result of tf.linalg.matmul
+        #     as it will have shape [a, 1];
+        # - lhs.shape == [batch] + [b]   , rhs.shape == [batch] + [b]
+        #   - in this case, we need to expand lhs to [1, b] and rhs to [b, 1],
+        #     and the resulting shape is (). We need to squeeze the result of
+        #     tf.linalg.matmul as it will have shape [1, 1].
+        squeeze_idxs = []
+        if lhs_dim - len(lhs_batch) == 1:
+          lhs = tf.expand_dims(lhs, lhs_dim - 1)
+          squeeze_idxs.append(len(lhs.shape) - 2)
+        if rhs_dim - len(rhs_batch) == 1:
+          rhs = tf.expand_dims(rhs, rhs_dim - 2)
+          squeeze_idxs.append(len(rhs.shape) - 1)
+        result = tf.linalg.matmul(lhs, rhs)
+        if len(squeeze_idxs) != 0:
+          result = tf.squeeze(result, squeeze_idxs)
+        return result
+
   new_id = iter(string.ascii_letters)
   lhs_axis_ids = [next(new_id) for _ in lhs.shape]
   rhs_axis_ids = [next(new_id) for _ in rhs.shape]
@@ -854,6 +1100,7 @@ def _select_and_gather_add(tangents: TfVal,
 
     # Unpacks the first element of a tuple.
     def fst(t):
+      assert t.dtype == double_word_dtype
       st = _shift_right_logical(t, const(double_word_dtype, nbits))
       return _bitcast_convert_type(
         _convert_element_type(st, word_dtype, double_word_dtype), dtype
@@ -919,7 +1166,7 @@ def _common_reduce_window(operand, init_val, reducer, window_dimensions,
                                    window_strides, padding, base_dilation,
                                    window_dilation)
 
-  o_spec = tf.TensorSpec(operand.shape, dtype=operand.dtype)
+  o_spec = tf.TensorSpec((), dtype=operand.dtype)
   reducer_fn = tf.function(reducer).get_concrete_function(o_spec, o_spec)
 
   if not isinstance(init_val, tf.Tensor):
@@ -1351,10 +1598,11 @@ tf_impl[lax_linalg.eig_p] = _eig
 
 def _eigh(operand: TfVal, lower: bool):
   if operand.shape[-1] == 0:
-    return tuple([operand, tf.reshape(operand, operand.shape[:-1])])
-  if not lower:
-    operand = tf.linalg.adjoint(operand)
-  w, v = tf.linalg.eigh(operand)
+    v, w = operand, tf.reshape(operand, operand.shape[:-1])
+  else:
+    if not lower:
+      operand = tf.linalg.adjoint(operand)
+    w, v = tf.linalg.eigh(operand)
   cast_type = { tf.complex64: tf.float32
               , tf.complex128: tf.float64 }.get(operand.dtype)
   if cast_type is not None:
