@@ -22,21 +22,22 @@ from ..core import Trace, Tracer
 from ..abstract_arrays import ShapedArray, raise_to_shaped
 from ..ad_util import add_jaxvals, add_jaxvals_p, zeros_like_jaxval, zeros_like_p
 from .. import linear_util as lu
-from ..util import unzip2, partial, safe_map, wrap_name, split_list
+from ..util import (unzip2, partial, safe_map, wrap_name, split_list,
+                    canonicalize_axis)
 from . import xla
 from . import partial_eval as pe
 
 map = safe_map
 
 
-def batch(fun : lu.WrappedFun, in_vals, in_dims, out_dim_dests):
+def batch(fun: lu.WrappedFun, in_vals, in_dims, out_dim_dests, axis_name):
   # executes a batched version of `fun` following out_dim_dests
-  batched_fun = batch_fun(fun, in_dims, out_dim_dests)
+  batched_fun = batch_fun(fun, in_dims, out_dim_dests, axis_name=axis_name)
   return batched_fun.call_wrapped(*in_vals)
 
 @lu.transformation_with_aux
-def batch_subtrace(master, in_dims, *in_vals, **params):
-  trace = BatchTrace(master, core.cur_sublevel())
+def batch_subtrace(main, in_dims, *in_vals, **params):
+  trace = BatchTrace(main, core.cur_sublevel())
   in_tracers = [BatchTracer(trace, val, dim) if dim is not None else val
                 for val, dim in zip(in_vals, in_dims)]
   outs = yield in_tracers, params
@@ -45,22 +46,28 @@ def batch_subtrace(master, in_dims, *in_vals, **params):
   yield out_vals, out_dims
 
 
-def batch_fun(fun : lu.WrappedFun, in_dims, out_dim_dests, sum_match=False):
+def batch_fun(fun : lu.WrappedFun, in_dims, out_dim_dests, axis_name,
+              sum_match=False):
   # transformation version of batch, which doesn't call the function
   fun, out_dims = batch_subtrace(fun)
-  return _batch_fun(fun, sum_match, in_dims, out_dims, out_dim_dests)
+  return _batch_fun(fun, axis_name, sum_match, in_dims, out_dims, out_dim_dests)
 
 @lu.transformation
-def _batch_fun(sum_match, in_dims, out_dims_thunk, out_dim_dests, *in_vals, **params):
+def _batch_fun(axis_name, sum_match, in_dims, out_dims_thunk, out_dim_dests,
+               *in_vals, **params):
   in_dims = in_dims() if callable(in_dims) else in_dims
+  in_dims = [
+    canonicalize_axis(dim, np.ndim(val)) if isinstance(dim, int) else dim
+    for val, dim in zip(in_vals, in_dims)]
   size, = {x.shape[d] for x, d in zip(in_vals, in_dims) if d is not not_mapped}
-  with core.new_master(BatchTrace) as master:
-    out_vals = yield (master, in_dims,) + in_vals, params
-    del master
+  with core.new_main(BatchTrace) as main:
+    with core.extend_axis_env(axis_name, size, main):
+      out_vals = yield (main, in_dims,) + in_vals, params
+    del main
   out_dim_dests = out_dim_dests() if callable(out_dim_dests) else out_dim_dests
   out_dims = out_dims_thunk()
   for od, od_dest in zip(out_dims, out_dim_dests):
-    if od is not None and not isinstance(od_dest, int) and not od_dest is last and not sum_match:
+    if od is not None and not isinstance(od_dest, int) and not sum_match:
       msg = f"vmap has mapped output but out_axes is {od_dest}"
       raise ValueError(msg)
   out_vals = map(partial(matchaxis, size, sum_match=sum_match), out_dims, out_dim_dests, out_vals)
@@ -73,9 +80,9 @@ def batch_fun2(fun : lu.WrappedFun, in_dims):
 
 @lu.transformation
 def _batch_fun2(in_dims, *in_vals, **params):
-  with core.new_master(BatchTrace) as master:
-    out_vals = yield (master, in_dims,) + in_vals, params
-    del master
+  with core.new_main(BatchTrace) as main:
+    out_vals = yield (main, in_dims,) + in_vals, params
+    del main
   yield out_vals
 
 
@@ -129,14 +136,29 @@ class BatchTrace(Trace):
     vals_in, dims_in = unzip2((t.val, t.batch_dim) for t in tracers)
     if all(bdim is not_mapped for bdim in dims_in):
       return primitive.bind(*vals_in, **params)
+    elif config.omnistaging_enabled and primitive in collective_rules:
+      axis_names = params['axis_name']
+      if not isinstance(axis_names, (tuple, list)):
+        axis_names = (axis_names,)
+      for i, axis_name in enumerate(axis_names):
+        frame = core.axis_frame(axis_name)
+        if frame.main_trace is not self.main:
+          continue
+        # We run the split_axis rule with tracers, which is supposed to never
+        # mix this axis name with another one. We will handle any invocations
+        # of collectives over the vmapped axis in a recursive call from a tracer.
+        if len(axis_names) > 1:
+          return split_axis(primitive, axis_name, tracers, params)
+        vals_out, dims_out = collective_rules[primitive](vals_in, dims_in, frame.size, **params)
+        results = map(partial(BatchTracer, self), vals_out, dims_out)
+        return results if primitive.multiple_results else results[0]
+    # TODO(mattjj,phawkins): if no rule implemented, could vmap-via-map here
+    batched_primitive = get_primitive_batcher(primitive)
+    val_out, dim_out = batched_primitive(vals_in, dims_in, **params)
+    if primitive.multiple_results:
+      return map(partial(BatchTracer, self), val_out, dim_out)
     else:
-      # TODO(mattjj,phawkins): if no rule implemented, could vmap-via-map here
-      batched_primitive = get_primitive_batcher(primitive)
-      val_out, dim_out = batched_primitive(vals_in, dims_in, **params)
-      if primitive.multiple_results:
-        return map(partial(BatchTracer, self), val_out, dim_out)
-      else:
-        return BatchTracer(self, val_out, dim_out)
+      return BatchTracer(self, val_out, dim_out)
 
   def process_call(self, call_primitive, f: lu.WrappedFun, tracers, params):
     assert call_primitive.multiple_results
@@ -145,15 +167,15 @@ class BatchTrace(Trace):
     if all(bdim is not_mapped for bdim in dims):
       return call_primitive.bind(f, *vals, **params)
     else:
-      f, dims_out = batch_subtrace(f, self.master, dims)
+      f, dims_out = batch_subtrace(f, self.main, dims)
       vals_out = call_primitive.bind(f, *vals, **params)
       return [BatchTracer(self, v, d) for v, d in zip(vals_out, dims_out())]
 
   def post_process_call(self, call_primitive, out_tracers, params):
     vals, dims = unzip2((t.val, t.batch_dim) for t in out_tracers)
-    master = self.master
+    main = self.main
     def todo(vals):
-      trace = BatchTrace(master, core.cur_sublevel())
+      trace = BatchTrace(main, core.cur_sublevel())
       return map(partial(BatchTracer, trace), vals, dims)
     return vals, todo
 
@@ -168,24 +190,24 @@ class BatchTrace(Trace):
               for x, d, mapped_invar in zip(vals, dims, mapped_invars)]
       dims = tuple(not_mapped if d is not_mapped else max(0, d - mapped_invar)
                    for d, mapped_invar in zip(dims, mapped_invars))
-      f, dims_out = batch_subtrace(f, self.master, dims)
+      f, dims_out = batch_subtrace(f, self.main, dims)
       vals_out = map_primitive.bind(f, *vals, **params)
       dims_out = tuple(d + 1 if d is not not_mapped else d for d in dims_out())
       return [BatchTracer(self, v, d) for v, d in zip(vals_out, dims_out)]
 
   def post_process_map(self, call_primitive, out_tracers, params):
     vals, dims = unzip2((t.val, t.batch_dim) for t in out_tracers)
-    master = self.master
+    main = self.main
     def todo(vals):
-      trace = BatchTrace(master, core.cur_sublevel())
+      trace = BatchTrace(main, core.cur_sublevel())
       return [BatchTracer(trace, v, d + 1 if d is not not_mapped else d)
               for v, d in zip(vals, dims)]
     return vals, todo
 
   def process_custom_jvp_call(self, prim, fun, jvp, tracers):
     in_vals, in_dims = unzip2((t.val, t.batch_dim) for t in tracers)
-    fun, out_dims1 = batch_subtrace(fun, self.master, in_dims)
-    jvp, out_dims2 = batch_custom_jvp_subtrace(jvp, self.master, in_dims)
+    fun, out_dims1 = batch_subtrace(fun, self.main, in_dims)
+    jvp, out_dims2 = batch_custom_jvp_subtrace(jvp, self.main, in_dims)
     out_vals = prim.bind(fun, jvp, *in_vals)
     fst, out_dims = lu.merge_linear_aux(out_dims1, out_dims2)
     if not fst:
@@ -195,9 +217,11 @@ class BatchTrace(Trace):
 
   def process_custom_vjp_call(self, prim, fun, fwd, bwd, tracers, *, out_trees):
     in_vals, in_dims = unzip2((t.val, t.batch_dim) for t in tracers)
-    fun, out_dims1 = batch_subtrace(fun, self.master, in_dims)
-    fwd, out_dims2 = batch_subtrace(fwd, self.master, in_dims)
-    bwd = batch_fun(bwd, out_dims2, in_dims, sum_match=True)
+    fun, out_dims1 = batch_subtrace(fun, self.main, in_dims)
+    fwd, out_dims2 = batch_subtrace(fwd, self.main, in_dims)
+    # TODO: Support collectives in custom_vjp?
+    bwd = batch_fun(bwd, out_dims2, in_dims,
+                    axis_name='__unused_axis_name', sum_match=True)
     out_vals = prim.bind(fun, fwd, bwd, *in_vals, out_trees=out_trees)
     fst, out_dims = lu.merge_linear_aux(out_dims1, out_dims2)
     if not fst:
@@ -297,14 +321,9 @@ defvectorized(xla.device_put_p)
 
 ### util
 
-class _Last(object): pass
-last = _Last()
-
 def broadcast(x, sz, axis):
   if core.get_aval(x) is core.abstract_unit:
     return core.unit
-  if axis is last:
-    axis = np.ndim(x)
   shape = list(np.shape(x))
   shape.insert(axis, sz)
   broadcast_dims = tuple(np.delete(np.arange(len(shape)), axis))
@@ -315,7 +334,8 @@ def moveaxis(x, src, dst):
     return core.unit
   if src == dst:
     return x
-  src, dst = src % x.ndim, dst % x.ndim
+  src = canonicalize_axis(src, x.ndim)
+  dst = canonicalize_axis(dst, x.ndim)
   perm = [i for i in range(np.ndim(x)) if i != src]
   perm.insert(dst, src)
   return x.transpose(perm)
@@ -327,10 +347,9 @@ def matchaxis(sz, src, dst, x, sum_match=False):
     return x
   elif type(src) == type(dst) == int:
     return moveaxis(x, src, dst)
-  elif type(src) == int and dst is last:
-    return moveaxis(x, src, -1)
   elif src is not_mapped and dst is not not_mapped:
-    return broadcast(x, sz, dst)
+    return broadcast(
+      x, sz, canonicalize_axis(dst, np.ndim(x) + 1))
   elif dst is None and sum_match:
     return x.sum(src)
   else:
@@ -356,21 +375,18 @@ def batch_jaxpr(jaxpr, size, batched, instantiate):
   f, batched_out = batched_traceable(f, size, batched, instantiate)
   avals_in = [_promote_aval_rank(size, a) if b else a
               for a, b in zip(jaxpr.in_avals, batched)]
-  in_pvals = [pe.PartialVal.unknown(aval) for aval in avals_in]
-  jaxpr_out, pvals_out, consts_out = pe.trace_to_jaxpr(f, in_pvals, instantiate=True)
-  avals_out, _ = unzip2(pvals_out)
-  jaxpr_out = core.TypedJaxpr(jaxpr_out, consts_out, avals_in, avals_out)
-  return jaxpr_out, batched_out()
+  jaxpr_out, _, consts = pe.trace_to_jaxpr_dynamic(f, avals_in)
+  return core.ClosedJaxpr(jaxpr_out, consts), batched_out()
 
 @lu.transformation_with_aux
 def batched_traceable(size, batched, instantiate, *vals):
   in_dims = [0 if b else None for b in batched]
-  with core.new_master(BatchTrace) as master:
-    trace = BatchTrace(master, core.cur_sublevel())
+  with core.new_main(BatchTrace) as main:
+    trace = BatchTrace(main, core.cur_sublevel())
     ans = yield map(partial(BatchTracer, trace), vals, in_dims), {}
     out_tracers = map(trace.full_raise, ans)
     out_vals, out_dims = unzip2((t.val, t.batch_dim) for t in out_tracers)
-    del master, out_tracers
+    del main, out_tracers
   if type(instantiate) is bool:
     instantiate = [instantiate] * len(out_vals)
   out_vals = [moveaxis(x, d, 0) if d is not not_mapped and d != 0
@@ -382,9 +398,9 @@ def batched_traceable(size, batched, instantiate, *vals):
 
 
 @lu.transformation_with_aux
-def batch_custom_jvp_subtrace(master, in_dims, *in_vals):
+def batch_custom_jvp_subtrace(main, in_dims, *in_vals):
   size, = {x.shape[d] for x, d in zip(in_vals, in_dims) if d is not not_mapped}
-  trace = BatchTrace(master, core.cur_sublevel())
+  trace = BatchTrace(main, core.cur_sublevel())
   in_tracers = [BatchTracer(trace, val, dim) if dim is not None else val
                 for val, dim in zip(in_vals, in_dims * 2)]
   outs = yield in_tracers, {}
@@ -408,8 +424,8 @@ def _merge_bdims(x, y):
     return x  # arbitrary
 
 
-@config.omnistaging_enablers.append
-def omnistaging_enabler() -> None:
+@config.register_omnistaging_disabler
+def omnistaging_disabler() -> None:
   global batch_jaxpr
 
   def batch_jaxpr(jaxpr, size, batched, instantiate):
@@ -417,6 +433,16 @@ def omnistaging_enabler() -> None:
     f, batched_out = batched_traceable(f, size, batched, instantiate)
     avals_in = [_promote_aval_rank(size, a) if b else a
                 for a, b in zip(jaxpr.in_avals, batched)]
-    jaxpr_out, avals_out, literals_out = pe.trace_to_jaxpr_dynamic(f, avals_in)
-    jaxpr_out = core.TypedJaxpr(jaxpr_out, literals_out, avals_in, avals_out)
-    return jaxpr_out, batched_out()
+    in_pvals = [pe.PartialVal.unknown(aval) for aval in avals_in]
+    jaxpr_out, pvals_out, consts_out = pe.trace_to_jaxpr(f, in_pvals, instantiate=True)
+    avals_out, _ = unzip2(pvals_out)
+    return core.ClosedJaxpr(jaxpr_out, consts_out), batched_out()
+
+
+# collective_rules can assume that the collective is only carried out throughout
+# the vmapped axis (i.e. no tuples in axis_name).
+collective_rules: Dict[core.Primitive, Callable] = {}
+split_axis_rules: Dict[core.Primitive, Callable] = {}
+
+def split_axis(primitive, split_name, args, params):
+  return split_axis_rules[primitive](split_name, args, params)
