@@ -171,22 +171,33 @@ def selu(x):
   scale = 1.0507009873554804934193349852946
   return scale * elu(x, alpha)
 
-def gelu(x):
+def gelu(x, approximate: bool = True):
   r"""Gaussian error linear unit activation function.
 
-  Computes the element-wise function:
+  If ``approximate=False``, computes the element-wise function:
+
+  .. math::
+    \mathrm{gelu}(x) = \frac{x}{2} \left(1 + \mathrm{erf} \left(
+      \frac{x}{\sqrt{2}} \right) \right)
+
+  If ``approximate=True``, uses the approximate formulation of GELU:
 
   .. math::
     \mathrm{gelu}(x) = \frac{x}{2} \left(1 + \mathrm{tanh} \left(
       \sqrt{\frac{2}{\pi}} \left(x + 0.044715 x^3 \right) \right) \right)
 
-  We explicitly use the approximation rather than the exact formulation for
-  speed. For more information, see `Gaussian Error Linear Units (GELUs)
+  For more information, see `Gaussian Error Linear Units (GELUs)
   <https://arxiv.org/abs/1606.08415>`_, section 2.
+
+  Args:
+    approximate: whether to use the approximate or exact formulation.
   """
-  sqrt_2_over_pi = np.sqrt(2 / np.pi).astype(x.dtype)
-  cdf = 0.5 * (1.0 + jnp.tanh(sqrt_2_over_pi * (x + 0.044715 * (x ** 3))))
-  return x * cdf
+  if approximate:
+    sqrt_2_over_pi = np.sqrt(2 / np.pi).astype(x.dtype)
+    cdf = 0.5 * (1.0 + jnp.tanh(sqrt_2_over_pi * (x + 0.044715 * (x ** 3))))
+    return x * cdf
+  else:
+    return jnp.array(x * (lax.erf(x / np.sqrt(2)) + 1) / 2, dtype=x.dtype)
 
 def glu(x, axis=-1):
   """Gated linear unit activation function."""
